# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" Simulation of a capacitively coupled plasma discharge: GUI """

# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

# Modules from the standard Python library
import math
from subprocess import run
from multiprocessing import Process, Pipe
import tkinter as tk
import tkinter.messagebox
import tkinter.scrolledtext

# Import required constants and parameters
from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *

# Import required modules, classes, and functions
from pysica.plasmapro.ccpla_init import *
from pysica.plasmapro.ccpla_print import *
from pysica.plasmapro.ccpla_plot import *
from pysica.plasmapro.ccpla_kernel import kernel
                
class CcplaWindow(tk.Frame):
    """ Create the main window """
        
    def __init__(self, charges, neutrals, ccp, parameters, options,
                 screen_width=SCREEEN_WIDTH, screen_height=SCREEEN_HEIGHT, master=None):

        # Class instances
        self.charges       = charges
        self.neutrals      = neutrals
        self.ccp           = ccp
        self.parameters    = parameters
        self.options       = options
        self.debug         = (self.options.debug_lev > 0)        

        # Create an ensamble of runtime plots and initialize them
        self.screen_width  = screen_width
        self.screen_height = screen_height
        self.runtime_plots = CcplaPlots(self.parameters, self.charges,
                                        self.ccp, self.screen_width,
                                        self.screen_height)

        # Normal variables
        self.sim_status         = False #True=running; False=stopped; None=paused
        self.error              = False
        #self.plot_counter       = 0
        #self.save_counter       = 0

        # Tk control variables
        self.dt_exp             = tk.IntVar()
        self.plot_delay         = tk.IntVar()
        self.plot_history       = tk.BooleanVar()
        self.plot_phase_space   = tk.BooleanVar()
        self.plot_field         = tk.BooleanVar()
        self.plot_distributions = tk.BooleanVar()
        self.plot_3D_positions  = tk.BooleanVar()

        # Activate all the plots by default
        self.set_all_plots(True)

        # Initialize Tk control variables 
        self.dt_exp.set(int(math.log10(self.parameters.dt_output)))
        self.plot_delay.set(DEF_PLOT_DELAY)

        # Create main window
        tk.Frame.__init__(self, master)

        # Add menu item "File"
        self.menubutton_file           = tk.Menubutton(self, text='File', justify=tk.LEFT)
        self.menubutton_file.grid(row=0, column=0, columnspan=1, sticky=tk.NW)

        self.menubutton_file.menu_file = tk.Menu(self.menubutton_file, tearoff=0)
        self.menubutton_file['menu']   = self.menubutton_file.menu_file
        #self.menubutton_file.menu_file.add_command(label='Change save directory name',
        #                                           command=self.change_basename)
        self.menubutton_file.menu_file.add_command(label='Reload configuration files',
                                                   command=self.reload_parameters)
        self.menubutton_file.menu_file.add_command(label='Edit configuration files',
                                                   command=self.edit_parameters)
        self.menubutton_file.menu_file.add_command(label='Quit' , command=self.exit_gui)

        # Add menu item "Parameters"
        self.menubutton_parameters = tk.Menubutton(self, text="Parameters", justify=tk.LEFT)
        self.menubutton_parameters.grid(row=0, column=1, columnspan=4, sticky=tk.NW)

        self.menubutton_parameters.menu_parameters = tk.Menu(self.menubutton_parameters,
                                                             tearoff=0)
        self.menubutton_parameters['menu'] = self.menubutton_parameters.menu_parameters
        self.menubutton_parameters.menu_parameters.add_command(
                label='Show physical parameters',
                command=lambda: self.show_parameters(physical=True))
        self.menubutton_parameters.menu_parameters.add_command(label='Show simulation parameters',
                                                               command=lambda: self.show_parameters(simulation=True))
        self.menubutton_parameters.menu_parameters.add_command(label='Show output parameters',
                                                               command=lambda: self.show_parameters(output=True))                
        self.menubutton_parameters.menu_parameters.add_command(label='Show gas properties', command=self.show_gases)
        self.menubutton_parameters.menu_parameters.add_separator()
        self.menubutton_parameters.menu_parameters.add_command(label='Show e-/neutral impact cross sections',
                                                               command=lambda: self.draw_cross_sections(electrons=True))
        self.menubutton_parameters.menu_parameters.add_command(label='Show ion/neutral impact cross sections',
                                                               command=lambda: self.draw_cross_sections(ions=True))  
        self.menubutton_parameters.menu_parameters.add_command(label='Show e-/ion recomb cross sections',
                                                               command=lambda: self.draw_cross_sections(recombination=True))
        self.menubutton_parameters.menu_parameters.add_separator()                
        self.menubutton_parameters.menu_parameters.add_command(label='Show e-/neutral impact parameters',
                                                               command=lambda: self.draw_cross_sections(electrons=True,
                                                                                                        plot_all=True))  
        self.menubutton_parameters.menu_parameters.add_command(label='Show ion/neutral impact parameters',
                                                               command=lambda: self.draw_cross_sections(ions=True, plot_all=True))
        self.menubutton_parameters.menu_parameters.add_command(label='Show e-/ion recombination parameters',
                                                               command=lambda: self.draw_cross_sections(recombination=True,
                                                                                                        plot_all=True))

        # Add menu item "Runtime plots"
        self.menubutton_runtime_plot = tk.Menubutton(self, text='Runtime Plots', justify=tk.LEFT)
        self.menubutton_runtime_plot.grid(row=0, column=5, columnspan=5, sticky=tk.NW)

        self.menubutton_runtime_plot.menu_runtime_plot = tk.Menu(self.menubutton_runtime_plot, tearoff=0)
        self.menubutton_runtime_plot['menu']   = self.menubutton_runtime_plot.menu_runtime_plot
        self.menubutton_runtime_plot.menu_runtime_plot.add_command(label='Select all',
                                                                    command = lambda: self.set_all_plots(True))
        self.menubutton_runtime_plot.menu_runtime_plot.add_command(label='Unselect all',
                                                                    command = lambda: self.set_all_plots(False))
        self.menubutton_runtime_plot.menu_runtime_plot.add_separator()
        self.menubutton_runtime_plot.menu_runtime_plot.add_checkbutton(label='Mean el energy and number vs time',
                                                                       variable=self.plot_history)
        self.menubutton_runtime_plot.menu_runtime_plot.add_checkbutton(label='Phase space plot',
                                                                       variable=self.plot_phase_space)
        self.menubutton_runtime_plot.menu_runtime_plot.add_checkbutton(label='Eletric field and charge',
                                                                       variable=self.plot_field)
        self.menubutton_runtime_plot.menu_runtime_plot.add_checkbutton(label='EEDF and IEDF',
                                                                       variable=self.plot_distributions)
        self.menubutton_runtime_plot.menu_runtime_plot.add_checkbutton(label='3D e- and ions positions',
                                                                       variable=self.plot_3D_positions)
        self.menubutton_runtime_plot.menu_runtime_plot.add_separator()        
        self.menubutton_runtime_plot.menu_runtime_plot.add_command(label='Plot list', command=self.show_plot_info)
        self.menubutton_runtime_plot.menu_runtime_plot.add_command(label='Plot export', command=self.print_plots)        

        # Add menu item "Help"
        self.menubutton_help           = tk.Menubutton(self, text='Help', justify=tk.LEFT)
        self.menubutton_help.grid(row=0, column=10, columnspan=1, sticky=tk.NW)

        self.menubutton_help.menu_help = tk.Menu(self.menubutton_help, tearoff=0)
        self.menubutton_help['menu']   = self.menubutton_help.menu_help
        self.menubutton_help.menu_help.add_command(label='About', command=self.show_about)

        # Add text region
        self.text_output = tk.Text(self,
                                   height=options.text_window_height,
                                   width=options.text_window_width,
                                   font=options.text_window_font,
                                   background=TEXT_WINDOW_BACK,
                                   foreground=TEXT_WINDOW_FORE)
        self.text_output.grid(row=1, column=0, columnspan=50)       

        # Add dt output scale
        self.dt_out_set = tk.Scale(self, from_=MIN_DT_OUTPUT_EXP, to=MAX_DT_OUTPUT_EXP, showvalue=True,#False,
                                   orient=tk.HORIZONTAL, length=300, 
                                   variable=self.dt_exp, command=self.set_dt_output)
        self.dt_out_set.grid(row=2, column=5, columnspan=10)

        # Add plot delay scale
        self.dt_out_set = tk.Scale(self, from_=MIN_PLOT_DELAY, to=MAX_PLOT_DELAY, resolution=RES_PLOT_DELAY, showvalue=False,
                                   orient=tk.HORIZONTAL, length=300, 
                                   variable=self.plot_delay, command=self.set_plot_delay)
        self.dt_out_set.grid(row=3, column=5, columnspan=10)


        # Add label to show dt_output
        self.dt_out_label = tk.Label(self, font = ('mono', 8 ), width=22, justify=tk.LEFT, anchor=tk.W)
        self.show_dt_out_label()

        # Add label to show paused/running
        self.status_label = tk.Label(self)
        self.show_status_label(text='Press RESET to start', color='red')

        # Add buttons
        self.button_reset = tk.Button(self, text='RESET', foreground='red', background='light yellow',
                                      command=self.reset_simulation)
        self.button_reset.grid(row=2, column=0, columnspan=2, sticky=tk.W)
        self.button_start = tk.Button(self, text='START', foreground='red', command=self.start_simulation)
        self.button_start.grid(row=2, column=35, columnspan=5)
        self.button_pause = tk.Button(self, text='Pause', foreground='blue', command=self.pause_restore_simulation)
        self.button_pause.grid(row=2, column=40, columnspan=5)
        self.button_stop = tk.Button(self, text='STOP', foreground='red', command=self.stop_simulation)
        self.button_stop.grid(row=2, column=45, columnspan=5)

        # Set status of buttons at program start
        self.button_reset["state"] = tk.NORMAL
        self.button_start["state"] = tk.DISABLED
        self.button_pause["state"] = tk.DISABLED
        self.button_stop["state"]  = tk.DISABLED                

        self.set_menu()

        self.grid()

        self.after(IDLE_TIME_START, self.show_about)

        self.initialize_kernel()

        
    # +-----------------------------------------------------+    
    # | Method to quit the main window and exit the program |
    # +-----------------------------------------------------+
    
    def exit_gui(self):
        """ Terminate the kernel and close the main window """
        self.quit_kernel()
        self.f_kernel.join()
        self.master.destroy() #self.quit()

        
    # +--------------------------------------+
    # | Methods to set the window properties |
    # +--------------------------------------+  
        
    def set_menu(self, change_names=True, edit_parameters=True, reload_parameters=True):
        index = self.menubutton_parameters.menu_parameters.index(tk.END)
        if self.parameters.isactive_recomb: self.state = tk.NORMAL
        else:                               self.state = tk.DISABLED
        self.menubutton_parameters.menu_parameters.entryconfigure(index,   state=self.state)
        self.menubutton_parameters.menu_parameters.entryconfigure(index-4, state=self.state)
        if change_names:      self.state = tk.NORMAL
        else:                 self.state = tk.DISABLED
        self.menubutton_file.menu_file.entryconfigure(0, state=self.state)         
        if edit_parameters:   self.state = tk.NORMAL
        else:                 self.state = tk.DISABLED
        self.menubutton_file.menu_file.entryconfigure(1, state=self.state)
        if reload_parameters: self.state = tk.NORMAL
        else:                 self.state = tk.DISABLED
        self.menubutton_file.menu_file.entryconfigure(2, state=self.state)
        

    # +-------------------------------------------+
    # | Methods to interact with kernel and pipes | 
    # +-------------------------------------------+    

    def initialize_kernel(self):
        """ Initialize the kernel process and the pipes """

        if self.debug: print('[GUI]   -> INITIALIZING KERNEL')
        
        (self.parent_signal, self.child_signal) = Pipe(duplex=True)
        (self.parent_time,   self.child_time)   = Pipe(duplex=True)        
        (self.parent_data,   self.child_data)   = Pipe(duplex=True)        
        self.f_kernel = Process(target=kernel,
                                args=(self.child_signal,
                                      self.child_time,
                                      self.child_data,
                                      (self.options.debug_lev >= KERNEL_DEBUG_LEV)
                                     )
                               )
        self.f_kernel.daemon = True
        self.f_kernel.start()
        
               
    def start_kernel(self):
        """ Tell the kernel to start iteration and send initializerd data to it """
        
        if self.debug: print('[GUI]   -> Sending start signal to kernel')
        self.parent_signal.send('start')
        if self.debug: print('[GUI]   -> Sending initialized data to kernel')
        self.parent_data.send((self.charges, self.neutrals, self.ccp, self.parameters, self.options))
        if self.debug: print('[GUI]   -> Sending requested simulation time to kernel')
        self.parent_time.send(self.parameters.dt_output)

    def stop_kernel(self):
        """ Flush pipes and tell the kernel to stop iteration """
        
        self.flush_pipes()
        self.parent_signal.send('stop')
        
    def quit_kernel(self):
        """ Tell kernel to quit, if doesn't, kill it """
        
        if self.debug: print('[GUI]   -> quitting kernel')
        self.show_status_label(text='Waiting for kernel to shutdown', color='red')
        self.parent_signal.send('quit')        
        self.after(IDLE_TIME_SHUTDOWN)
        if self.f_kernel.is_alive():
            self.show_status_label(text='Forcing kernel to shut down...', color='red')
            if self.debug: print('[GUI]   -> terminating kernel')            
            self.f_kernel.terminate()
            self.after(IDLE_TIME_SHUTDOWN)
            if self.f_kernel.is_alive():
                if self.debug: print('[GUI]   -> killing kernel')
                self.f_kernel.kill()
        if self.debug: print('[GUI]   -> joining kernel')
        self.f_kernel.join()        


    def flush_pipes(self, flush_signal=True, flush_time=True, flush_data=True):
        """ Read any residual data from the pipes """
        
        if self.debug: print('[GUI]   -> CHECKING PIPES')
        if flush_signal:
            while self.parent_signal.poll():
                if self.debug: print('[GUI]   -> FLUSHING SIGNAL PIPE')
                signal = self.parent_signal.recv()
        if flush_time:    
            while self.parent_time.poll():
                if self.debug: print('[GUI]   -> FLUSHING TIME PIPE')                
                time = self.parent_time.recv()
        if flush_data:
            while self.parent_data.poll():
                if self.debug: print('[GUI]   -> FLUSHING DATA PIPE')                
                (self.charges, self.neutrals, self.ccp, self.parameters, self.options) = self.parent_data.recv()

        
    def close_pipes(self):
        """ Close the pipes """
        if self.debug: print('[GUI]   -> CLOSING PIPES')
        
        self.parent_signal.close()
        self.parent_time.close()
        self.parent_data.close()

        
    # +------------------------------------------------+
    # | Main iteration and methods to interact with it |
    # +------------------------------------------------+    
        
    def iteration(self):
        """ Main simulation iteration """

        # If self.sim_status if False, the simulation must end: exit from simulation loop
        if (self.sim_status is False):
            return        
        # If self.sim_status if True or None, the simulation is running or paused
        else:            
            # Check if the kernel has new calculated data
            self.data_ready = self.parent_signal.poll()
            if self.data_ready:
                self.show_status_label(text='Kernel is ready  ', color='green')
                if (self.sim_status is None): self.button_stop["text"] = 'STOP'          
            else:
                self.show_status_label(text='Kernel is busy ...', color='red')
                if (self.sim_status is None): self.button_stop["text"] = 'KILL'
            self.update_idletasks()
            # If there is new data and simulation is not paused
            if (self.data_ready and (self.sim_status is True)):
                message = self.parent_signal.recv()
                if self.debug: print('[GUI]   -> Message from kernel: ' + message)# + '\n')
                if self.debug: print('[GUI]   -> Reading data from kernel')
                (self.charges, self.neutrals, self.ccp, self.parameters, self.options) = self.parent_data.recv()
                if self.debug: print('[GUI]   -> Data read from kernel')                
                # Syncronize requested iteration duration with window selector
                self.parameters.dt_output = 10.0**int(self.dt_exp.get())                
                # Update data on text window
                self.show_runtime_info()                
                # Update plots 
                self.draw_plots()                
                # Save data to files
                if (self.parameters.save_delay > 0):
                    self.save_counter += 1
                    if (self.save_counter >= self.parameters.save_delay):
                        if self.debug: print('[GUI]   -> Saving data')                                        
                        self.charges.save_data_to_files()
                        self.neutrals.save_data_to_files(self.charges.time)
                        self.ccp.save_data_to_files(self.charges.time)
                        if self.debug: print('[GUI]   -> Data saved')
                        self.save_counter = 0                    
                # Save actual time to be shown in next iteration
                self.time_before          = self.charges.time
                self.n_active_el_before   = self.charges.n_active(0)
                self.electric_bias_before = self.ccp.V
                if (self.charges.n_active(0) <= 0):
                    self.show_end_simulation(text ='\n\nSimulation interrupted \n (no more electrons)\n\n',
                                             color='red')
                    self.stop_simulation(confirm_stop=False)
                elif ( (self.parameters.sim_duration > 0) and (self.charges.time >= self.parameters.sim_duration) ):
                    self.show_end_simulation(text='\n\nSimulation completed\n\n', color='blue')
                    self.stop_simulation(confirm_stop=False)
                else:
                    self.parent_signal.send('continue')
                    self.parent_time.send(self.parameters.dt_output)                    
        # Show the simulation output time
        self.show_dt_out_label()       
        # Before repeating the loop, update widgets and wait some time to allow responsivity of the window
        self.update_idletasks()
        self.after(IDLE_TIME, self.iteration)
        
                
    def start_simulation(self):
        """ Begin a new simulation run """

        self.plot_counter = 0
        self.time_before          = self.charges.time
        self.n_active_el_before   = self.charges.n_active(0)
        self.electric_bias_before = self.ccp.V      
        
        self.draw_plots(force_plot=True)       

        self.set_menu(change_names=False, edit_parameters=False, reload_parameters=False)
        self.button_start["state"] = tk.DISABLED
        self.button_pause["state"] = tk.NORMAL
        self.button_stop["state"]  = tk.DISABLED         
        self.show_runtime_info()
        
        self.update_idletasks()
         
        self.sim_status = True
        if self.debug: print('[GUI]   -> start kernel')
        self.start_kernel()
        if self.debug: print('[GUI]   -> start iteration')
        self.after(IDLE_TIME, self.iteration)

        
    def pause_restore_simulation(self):
        """ Pause the simulation or restore it from paused state """
        
        if (self.sim_status == True):
            self.sim_status = None
            self.button_stop["state"] = tk.NORMAL
            self.button_pause["text"] = 'Continue'
        elif (self.sim_status == None):
            self.sim_status = True
            self.button_stop["state"] = tk.DISABLED
            self.button_pause["text"] = 'Pause'
        if self.debug: print('[GUI]   -> sim_status = ' + str(self.sim_status))
                

    def stop_simulation(self, confirm_stop=True):
        """ End the running simulation """
        
        if self.data_ready:
            if confirm_stop:
                self.answer = tkinter.messagebox.askyesno( title='Stop simulation',
                                                           message=('Stop the simulation ?'),
                                                           icon='question')
                if not self.answer: return
            self.stop_kernel()            
        else:
            self.answer = tkinter.messagebox.askyesno( title='WARNING',
                                                       message=('WARNING! Killing the running kernel may produce data loss \n'
                                                                + 'Do you want to proceed ?'),
                                                       icon='warning')
            if not self.answer: return
            else:
                self.quit_kernel()
                self.close_pipes()
                self.initialize_kernel()
        #self.flush_pipes()
        self.sim_status = False        
        self.button_pause["state"]  = tk.DISABLED
        self.button_stop["text"]    = 'STOP'
        self.button_stop["state"]   = tk.DISABLED
        self.show_status_label(text='Press RESET to start over', color='blue')
        self.button_reset["state"]  = tk.NORMAL
        self.set_menu()       


    def reset_simulation(self):
        """ Reset all data for a new simulation run """
        
        # Initialize particle data
        if self.debug: print('[GUI]   -> Initializing particle data')        
        (status, message) = initialize_ensambles(self.charges, self.neutrals, self.parameters, self.options)
        if (status != 0):
            self.show_message(message, fore='red', back='light yellow')
            return
        self.parameters.dt_output = 10.0**int(self.dt_exp.get())

        # Initialize plots
        if self.debug: print('[GUI]   -> Initializing plots')        
        #self.runtime_plots.initialize_graphs(reset=True)
        self.runtime_plots.reset_graphs(self.ccp, self.parameters)
        self.runtime_plots.clear_graphs()
        self.runtime_plots.reset_history()
        self.draw_plots(force_plot=True)                    
        self.after(IDLE_TIME_GRAPHS)        
        self.runtime_plots.lower_graphs()
        self.plot_counter = 0

        # Create save directory and initialize savefiles
        if (self.parameters.save_delay > 0):
            if self.debug: print('[GUI]   -> Initializing save directory')
            initialize_filenames(self.neutrals, self.parameters)                        
            self.answer = tkinter.messagebox.askyesno( title='Check',
                                                       message=('Use this name for the save directory? \"'
                                                                + str(self.parameters.basename)
                                                                + '\"') )
            if not self.answer: self.change_basename()
            create_save_dir(self.parameters)
            if self.debug: print('[GUI]   -> Initializing savefiles')
            self.charges.initialize_savefiles(self.parameters.filename_stat_ele,
                                              self.parameters.filename_distrib_ele,
                                              self.parameters.filename_distrib_ion,
                                              append=False, sep=SEP, ext=EXT)
            self.neutrals.initialize_savefile(self.parameters.filename_stat_neu,
                                              append=False, sep=SEP, ext=EXT)
            self.ccp.initialize_savefiles(self.parameters.filename_I, self.parameters.filename_V,
                                          append=False, sep=SEP, ext=EXT)            
            self.save_counter = 0          
            
        # Arrange flags, windows and buttons status
        if self.debug: print('[GUI]   -> Initializing status')  
        self.sim_status = False
        self.show_runtime_info(erase_only=True)
        self.show_status_label(text='Press START to begin simulation', color='red')
        self.button_reset["state"] = tk.DISABLED
        self.button_pause["text"]  = 'Pause'
        self.button_pause["state"] = tk.DISABLED
        self.button_stop["text"]   = 'STOP'
        self.button_stop["state"]  = tk.DISABLED
        self.button_start["state"] = tk.NORMAL

        self.update_idletasks()
        


    # +-----------------------------------------------+
    # | Metods to interact with simulation parameters |
    # +-----------------------------------------------+

    def edit_parameters(self):
        try:
            #editor_process = run([EDITOR_NAME, FILENAME_CONFIG, FILENAME_NEUTRALS])
            editor_process = run([EDITOR_NAME, FILENAME_CONFIG])
        except FileNotFoundError:
            string = ('Error while trying to execute command: \"'
                      + EDITOR_NAME + '\" '
                      + FILENAME_CONFIG
#                      + ' '
#                      + FILENAME_NEUTRALS
                      )
            self.show_message(message=string, fore='red', back='light yellow')
            return
        self.reload_parameters()
        
                        
    def reload_parameters(self): 
        self.error   = False
        self.warning = False
        ERRORS = ''
        # Reload parameters from configuration file
        (status, message) = initialize_parameters(self.parameters,
                                                  verbose=False,
                                                  saveonly=False,
                                                  filename_config=FILENAME_CONFIG,
                                                  filename_defaults=FILENAME_DEFAULTS,
                                                  restricted=True)
        if (status != 0):
            self.error= True
            ERRORS += message + '\n' 
        # Recofigure reactor parameters
        self.ccp.__init__(self.parameters.distance,
                          self.parameters.length,
                          self.parameters.V_bias,
                          self.parameters.frequency,
                          self.parameters.phase,
                          self.parameters.N_cells,
                          self.parameters.lateral_loss)
        
        # Reconfigure neutral particle properties
#        self.neutrals.__init__(self.parameters.N_sigma,
#                               self.parameters.N_sigma_ions,
#                               self.parameters.T_neutrals,
#                               self.parameters.p_neutrals,
#                               self.parameters.min_scattered,
#                               self.parameters.isactive_recomb,
#                               filename=FILENAME_NEUTRALS)
#        (status, message) = self.neutrals.read_error
#        if (status !=0):
#            self.error= True
#            ERRORS += 'Error in file \"' + FILENAME_NEUTRALS + '\": ' + message + '\n'
#        # Read neutral properties from file
#        (status, message) = self.neutrals.read_properties(FILENAME_NEUTRALS,
#                                                          SEP, 
#                                                          self.parameters.e_min_sigma,
#                                                          self.parameters.e_max_sigma,
#                                                          self.parameters.e_min_sigma_ions,
#                                                          self.parameters.e_max_sigma_ions)
#        if (status !=0):
#            self.error= True
#            ERRORS += 'Error in file \"' + FILENAME_NEUTRALS + '\": ' + message + '\n'

#        if not self.error:
#            # Reload cross sections and calculate other impact parameters
#            (status, message) = initialize_cross_sections(self.neutrals, self.options, self.parameters.isactive_recomb)
#            if (status !=0):
#                self.error= True
#                ERRORS += message + '\n'
        if self.error:
            self.button_reset["state"] = tk.DISABLED
            self.button_start["state"] = tk.DISABLED
            self.show_message('Errors encountered while loading parameters \n\n' + ERRORS,
                              fore='red', back='light yellow')
            self.show_status_label(text='Errors: check configuration files', color='red')
        else:
            # Reconfigure electron and ion properties
            #self.charges.__init__(self.neutrals.types+1, self.parameters.Nmax_particles, START_WEIGHT,
            #                      self.parameters.rescale_factor)            
            initialize_ensambles(self.charges, self.neutrals, self.parameters, self.options)            
            self.show_message('Parameters reloaded successfully', fore='blue')            
            self.button_reset["state"] = tk.NORMAL
            self.button_start["state"] = tk.DISABLED
            self.set_menu()
            self.set_dt_output()
            self.show_status_label(text='Press RESET to start over', color='blue')
                        
    def set_dt_output(self, set_value=None):
        # If set_value is None, self.parameters.dt_output is assumed to have been already set
        if (set_value is not None): self.parameters.dt_output = 10.0**int(set_value)
        # Check that output periodicity is not smaller than timestep
        if ( (not self.parameters.dt_var) and (self.parameters.dt_output < self.charges.dt) ):
            self.parameters.dt_output = self.charges.dt
        self.dt_exp.set(int(math.log10(self.parameters.dt_output)))
        self.show_dt_out_label()

    def set_plot_delay(self, set_value):
        self.plot_counter = 0
        self.show_dt_out_label()                

    def set_all_plots(self, state=True):
        self.plot_history.set(state)
        self.plot_phase_space.set(state)
        self.plot_field.set(state)
        self.plot_distributions.set(state)
        self.plot_3D_positions.set(state)


    # +--------------------------------------+
    # | Methods to deal with data file names |
    # +--------------------------------------+  
        
    def change_basename(self):
        self.filenames_window = tk.Toplevel(self)
        self.filenames_window.title('Enter save directory name')
        self.filenames_window.resizable(False,False)
        self.filenames_window.message = tk.Entry(self.filenames_window, 
                                                 font=FONT_MESSAGE, foreground='blue',takefocus=True,
                                                 width=40,
                                                 justify=tk.LEFT, relief=tk.RIDGE)
        self.filenames_window.message.insert('0', self.parameters.basename)
        self.filenames_window.message.grid()
        self.filenames_window.button_ok = tk.Button(self.filenames_window, text="Ok", command=self.get_basename)
        self.filenames_window.button_ok.grid()
        self.filenames_window.message.focus_set()
        self.filenames_window.grab_set()
        self.wait_window(self.filenames_window)

    def get_basename(self):
        self.parameters.basename = self.filenames_window.message.get()
        initialize_filenames(self.neutrals, self.parameters, self.parameters.basename)
        self.filenames_window.destroy()

        
    # +------------------------------------------------+
    # | Methods to show the information on the windows |
    # +------------------------------------------------+        
    
    def show_dt_out_label(self):
        if (self.charges.dt == 0): self.dt_print = 'variable'
        else:                      self.dt_print = unit_manager.print_unit(self.charges.dt,'s', 1)
        # This is necessay since scale widget with resolution > 1  and minimum 1 will give zero (not 1) as minimum value
        if (self.plot_delay.get() == 0): self.dt_plot = self.parameters.dt_output
        else:                            self.dt_plot = self.plot_delay.get() * self.parameters.dt_output                
        self.dt_out_label["text"] = 'Output every: ' + unit_manager.print_unit(self.parameters.dt_output,'s', 1) + '\n' +\
                                    'Plot every:   ' + unit_manager.print_unit(self.dt_plot, 's', 1) + '\n' +\
                                    'Timestep:     ' + self.dt_print
        self.dt_out_label.grid(row=2, column=15, columnspan=8) 
              
    def show_status_label(self, text, color):
        self.status_label["text"]       = text
        self.status_label["foreground"] = color
        self.status_label.grid(row=2, column=22, columnspan=12)
                
    def show_about(self):
        self.about_window = tk.Toplevel(self)
        self.about_window.title('Copyright Info')
        self.about_window.resizable(False,False)
        self.about_window.message = tk.Message(self.about_window, text=GPL_MESSAGE, justify=tk.CENTER, relief=tk.RIDGE)
        self.about_window.message.grid()
        self.about_window.button  = tk.Button(self.about_window, text="Dismiss", command=self.about_window.destroy)
        self.about_window.button.grid()
        self.about_window.attributes("-topmost", True)
        #self.text_output.delete('0.0',tk.END)
        #self.text_output.insert('0.0', GPL_MESSAGE)
        #self.text_output.grid(row=1, column=0, columnspan=50)         
                        
    def show_gases(self):
        self.gases_window = tk.Toplevel(self)
        self.gases_window.title('Gas Properties')
        self.gases_window.resizable(False,False)
        self.gases_window.message = tk.Message(self.gases_window, text=print_gas_information(self.neutrals),
                                               font=FONT_MESSAGE_SMALL, foreground='blue', width=1500,
                                               justify=tk.LEFT, relief=tk.RIDGE)
        self.gases_window.message.grid()
        self.gases_window.button  = tk.Button(self.gases_window, text="Dismiss", command=self.gases_window.destroy)
        self.gases_window.button.grid()
                
    def show_parameters(self, physical=False, simulation=False, output=False, filenames=False):
        self.parameters_window = tk.Toplevel(self)
        self.parameters_window.title('Parameters')
        self.parameters_window.resizable(False,False)
        self.parameters_window.message = tk.Message(self.parameters_window,
                                                    text=print_simulation_information(self.neutrals, self.ccp,
                                                                                      self.parameters, self.options,
                                                                                      print_physical=physical,
                                                                                      print_simulation=simulation,
                                                                                      print_output=output,
                                                                                      print_filenames=filenames),
                                                    font=FONT_MESSAGE_SMALL, foreground='blue', width=1500,
                                                    justify=tk.LEFT, relief=tk.RIDGE)
        self.parameters_window.message.grid()
        self.parameters_window.button  = tk.Button(self.parameters_window, text="Dismiss",
                                                   command=self.parameters_window.destroy)
        self.parameters_window.button.grid()
        
    def show_plot_info(self):
        try:
            string = self.runtime_plots.graph_list(expanded=True)
            if (string is None): string = 'No plots'
        except AttributeError:
            string = 'No plots'
        self.plotinfo_window = tk.Toplevel(self)
        self.plotinfo_window.title('Plot Windows Properties')
        self.plotinfo_window.resizable(False,False)
        self.plotinfo_window.text = tk.scrolledtext.ScrolledText(self.plotinfo_window,
                                                                 font=FONT_MESSAGE_SMALL,
                                                                 background='light grey',
                                                                 foreground='blue',
                                                                 width=90,
                                                                 height=30)
        self.plotinfo_window.text.grid()
        self.plotinfo_window.text.insert(tk.INSERT, string)
        self.plotinfo_window.text.configure(state='disabled')
        self.plotinfo_window.button  = tk.Button(self.plotinfo_window, text="Dismiss",
                                                   command=self.plotinfo_window.destroy)
        self.plotinfo_window.button.grid()
            
    def show_runtime_info(self, erase_only=False):
        if erase_only:
            self.text_output.delete('0.0',tk.END)
        else:
            self.text_output.delete('0.0',tk.END)
            self.text_output.insert('0.0',print_runtime_info(self.charges,
                                                             self.neutrals,
                                                             self.ccp,
                                                             self.parameters,
                                                             self.options,
                                                             self.time_before,
                                                             self.electric_bias_before,
                                                             self.n_active_el_before)
                                        )
        self.text_output.grid(row=1, column=0, columnspan=50)
                
    def show_end_simulation(self, text='End simulation', color='black'):
        self.end_simulation_window = tk.Toplevel(self) 
        self.end_simulation_window.title('Ccpla End')
        self.end_simulation_window.resizable(False,False)
        self.end_simulation_window.message = tk.Message(self.end_simulation_window,
                                                        text=text, font=FONT_MESSAGE_BIG,
                                                        foreground=color,
                                                        width=1500,
                                                        justify=tk.CENTER, relief=tk.RIDGE)
        self.end_simulation_window.message.grid()
        self.end_simulation_window.button = tk.Button(self.end_simulation_window, text="Dismiss",
                                                      command=self.end_simulation_window.destroy)
        self.end_simulation_window.button.grid()
        self.end_simulation_window.attributes("-topmost", True)       
                        
    def show_message(self, message, fore='black', back='light grey'):
        self.message_window = tk.Toplevel(self)
        self.message_window.title('Message')
        self.message_window.resizable(False,False)
        self.message_window.message = tk.Message(self.message_window,
                                                 text=message,
                                                 background=back,
                                                 foreground=fore,
                                                 width=1500,
                                                 font=self.options.text_window_font,
                                                 justify=tk.CENTER,
                                                 relief=tk.RIDGE)
        self.message_window.message.grid()
        self.message_window.button  = tk.Button(self.message_window, text="Dismiss", command=self.message_window.destroy)
        self.message_window.button.grid()
        self.message_window.attributes("-topmost", True)

    def draw_cross_sections(self, electrons=False, ions=False, recombination=False, plot_all=False):
        (status, message) = plot_cross_sections(self.neutrals,
                                                electrons=electrons,
                                                ions=ions,
                                                recombination=recombination,
                                                plot_all=plot_all,
                                                dot_points=self.parameters.dot_points) 
        if (status !=0): self.show_message(message, fore='red', back='light yellow')

    def draw_plots(self, force_plot=False):
        self.runtime_plots.update_history(self.charges, self.charges.time)
        if ( force_plot or (self.plot_counter >= self.plot_delay.get()) ):
            self.plot_counter = 0
            if self.plot_history.get():
                self.runtime_plots.plot_n_and_energy_data(self.charges.time)
                self.update_idletasks()
            if self.plot_phase_space.get():
                self.runtime_plots.plot_phase_space_data(self.charges.time, self.charges)
                self.update_idletasks()
            if self.plot_field.get():
                self.runtime_plots.plot_potential(self.charges.time, self.ccp)
                self.update_idletasks()
            if self.plot_distributions.get():
                self.runtime_plots.plot_distributions(self.charges.time, self.charges)
                self.update_idletasks()
            if self.plot_3D_positions.get():
                self.runtime_plots.plot_positions(self.charges.time, self.charges)
                self.update_idletasks()
        self.plot_counter += 1

    def print_plots(self):
        self.runtime_plots.print_graphs('png', SAVE_DIRECTORY_NAME)
    

def main(charges, neutrals, ccp, parameters, options):
    """ Graphical interface for ccpla script """

    root = tk.Tk()
    root.title("Ccpla GUI")
    root.resizable(False, False)
        
    # Get the screen dimensions
    screen_width  = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()
       
    # Main window
    main_window = CcplaWindow(charges, neutrals, ccp, parameters, options, screen_width, screen_height, master=root)

    root.mainloop()

              



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Tools to import/export numerical data from/to data files in ASCII format

In this module some classes and methods are provided to deal with 
    ASCII data files of different types.

    * Configuration files, containing the values to assign 
      to some variables, in the form:
          # Comment
          variable1 = <value1> #comment
          <empty lines>
          variable2 = <value2> #comment
      empty lines and comments starting with '#' are allowed and ignored.

      Data are registered in a dictionary: 
          { 'variable1': [<value1>], 'variable2': [<value2>], ... }

    * Data sequence, containing a sequence of ordered values, 
      in the following format:
          value1  <separator> value2  <separator> value3 ... <separator> valueN

      Almost any type of ASCII character sequence can be used as separator, 
      but usually EOL, "," or TAB are more convenient.

      Space is *not* usable as separation character, since spaces are 
      removed before text parsing

    * Data grids, i.e. two dimensional matrix of data, expressed in ASCII format
          #Comment
          #Comment
          <Value11> <separator> <Value12> <separator> <Value13>   #Comment
          <Value21> <separator> <Value22> <separator> <Value23>
      empty lines and comments starting with '#' are allowed and ignored.

      Data are stored in a 2-d numpy array, just like the following
          [ [<Value11>, <Value12>, <Value13>], 
            [<Value21>, <Value22>, <Value23>] ]
      or the following 
          [ [<Value11>, <Value21>], 
            [<Value12>, <Value22>], 
            [<Value13>, <Value23>] ]

      You may also specify a number of lines to skip at the beginning 
      of the ASCII file: this is often useful in order to skip the header 
      of some data file formats.

    Documentation is also available in the docstrings.
"""


import numpy

from ..parameters import *

#+--------------------+
#| Configuration Data |
#+--------------------+

class ConfigurationData:
    """ Configuration data.
 
        This class defines a dictionary used to store variable names 
        and associated values: typical use is reading the values of 
        some variables from a configuration file.

        Format of the configuration file is like the following  
            #Comment
            variable1 = 123.45E-6
            variable2 = 7654.3E-2  #Comment
        empty lines and comments starting with '#' are allowed and ignored.
                
        The data are stored in a dictionary with the following form: 
            { <variable>: [list] }, where 
            * <variable> -> string describing a variable of some kind 
            * [list]     -> list contaning the variable value, 
                            and optionally other information
                            (see docstring of method read_file())
    """

    def __init__(self):
        """ Creates a dictionary for configuration data.

            Creates an empty dictionary, which will contain the couples 
            { <variable-name>: [list] }.
            To fill the dictionary with values read from a ascii file, 
            use the read_file() method.

            Initialized data attributes
            ---------------------------

            self.d:         an empty dictionary
        """

        # Dictionary in which variable names and corresponding
        # default values must be stored
        self.d = {}     

        
    def read_file(self, filename, check_values=False):
        """ Read configuration data from a file.

            This method reads variable names and values from 
            a ASCII file and stores them in the dictionary self.d 
            (which was created at the instance creation).

            It is also possible to ask the method to check 
            variable names and values and return an
            error if they do not meet some constraints, 
            by setting to True the parameter check_values.
            In this case, *before* calling the method, 
            you must fill the dictionary self.d 
            with couples of the type 
                { <variable1>: [list1], <variable2>: [list2], ... }
            where each list must have the format: 
                [ <value>, <string>, <code>, <value1>, <value2> ]
            * <value>:  numeric value associated to the variable, 
                        which will be overwritten if the ASCII file 
                        contains a different value for that variable
                        typically, this is the default value
            * <string>: a string describing the variable meaning
            * <code>:   a string describing the allowed range 
                        for <value>
                        'checkmin':            <value> >= <value1>
                        'checkmax':            <value> <= <value1>
                        'checkminstrict':      <value> >  <value1>
                        'checkmaxstrict':      <value> <  <value1>
                        'checkrange':          <value1> <= <value> <= <value2>
                        'checkrangestrict':    <value1> <  <value> <  <value2>
                        'checkrangestrictmin': <value1> <  <value> <= <value2>
                        'checkrangestrictmax': <value1> <= <value> <  <value2>
                        'boolean':             <value> = 0 or 1
            * <value1>: minimum, maximum, or lower boundary of allowed range 
            * <value2>: upper boundary of allowed range

            Example:
                particle = ConfigurationData()
                particle.dic = {        
                    "position":     [ 1.0, 'particle position' ] , 
                    "velocity":     [ 2.0, 'particle velocity', 
                                      'checkmin', 0.0 ], 

                    "energy":       [ 3.0, 'particle acceleration, ']
                    "mass":         [ 1.0, 'particle mass', 'checkrange', 
                                      1.0, 40.0 ],
                    "active":       [ 1, 'active particle', boolean ]
                                }
                particle.read_file("particledata.dat", check_values=True)


            Format of the ASCII file is like the following                     
                #Comment
                variable1 = 123.45E-6
                variable2 = 7654.3E-2  #Comment

            Parameters
            ----------

            filename:     a string containing the name of input data file
            check_values: if set to True, checks that variable names in file 
                          are already present in dictionary
                          and that their values are inside specified ranges

            Initialized data attributes
            ---------------------------

            self.d: dictionary containing variables and associated values

            Returns
            -------

            (status, message)
            status:  0 = no error
                     1 = could not open file
                     2 = syntax error
                     3 = invalid variable assignment
                     4 = unknown variable
                     5 = bad range specification
                     6 = variable out of allowed range
                     7 = bad keyword in range specification
            message: a string containing an error message or 'Ok'
        """

        # These variables are used for error checking            
        status   = 0
        message  = OK

        try:
            f_input = open(filename, 'r')
        except IOError:
            status  = 1
            message = 'could not open file'
            return (status, message)

        # Read command lines from input file
        line = 0
        for string in f_input:
            line = line + 1
            # Cut newline charater from the end of command line
            string = string.strip(EOL)
            # Cut line portion at right of '#' char
            string = string.split('#')[0]
            # Cut spaces from beginning and end of command line
            string = string.strip()
            # Ignore line if it is empty
            if (len(string) != 0):    
                # Separate variable name from the value we want to assign to it
                l = string.split("=")   
                # If line doesn't contain "=" character, we have a syntax error
                if len(l) < 2:
                    status = 2
                    message = "line "+str(line)+": invalid syntax" 
                    break
                left_string  = l[0].strip()     # Variable name
                right_string = l[1].strip()     # Variable value

                # Try to transform right part of assignemnt to floating-point
                try:
                    value = float(right_string)
                except ValueError:
                    status = 3
                    message = ("line "+str(line)
                               + ": assignment to variable \""
                               + left_string + "\" is not a number")
                    break

                # Try to recognise variable name among dictionary
                if (check_values):      
                    check = False
                    for (v, l) in self.d.items():
                        if (left_string == v): 
                            self.d[left_string][0] = value
                            check = True
                            break                                   
                    if not check:
                        status = 4
                        message = ("line " + str(line)
                                   + ": unknown variable \"" + left_string
                                   + "\"")
                        break

                # If check is not required, just add the (variable: [value]) couple to dictionary
                else:
                    self.d[left_string] = [value]

                # Check if value is in the valid range
                # (if check was requested and range was specified)
                if (check_values & (len(self.d[left_string]) > 2)):
                    l = self.d[left_string]
                    e5 = ('variable \'' + left_string
                          + '\': bad specification of range boundaries')
                    e6 = l[1] + ' is out of allowed range '
                    e7 = ('variable \'' + left_string
                          + '\': unknown range specification keyword')
                    if (l[2] == 'checkmin'):
                        if (len(l) < 4):
                            status = 5
                            message = e5
                            break                                                   
                        elif (l[0] < l[3]): #value < min
                            status = 6
                            message = e6 + '( '+left_string+' >= '+str(l[3])+' )'
                            break
                    elif (l[2] == 'checkminstrict'):
                        if (len(l) < 4):
                            status = 5
                            message = e5
                            break                                                   
                        elif (l[0] <= l[3]): #value <= min
                            status = 6
                            message = e6 + '( '+left_string+' > '+str(l[3])+' )'
                            break
                    elif (l[2] == 'checkmax'):
                        if (len(l) < 4):
                            status = 5
                            message = e5
                            break                                                   
                        elif (l[0] > l[3]):                               #value > max
                            status = 6
                            message = e6 + '( '+left_string+' <= '+str(l[3])+' )'
                            break
                    elif (l[2] == 'checkmaxstrict'):
                        if (len(l)<4):
                            status = 5
                            message = e5
                            break                                                   
                        elif (l[0] >= l[3]):                      #value >= max
                            status = 6
                            message = e6 + '( '+left_string+' < '+str(l[3])+' )'
                            break
                    elif (l[2] == 'checkrange'):
                        if (len(l) < 5):
                            status = 5
                            message = e5
                            break                                                   
                        elif ((l[0] < l[3]) | (l[0] > l[4])):     #value<min or value>max
                            status = 6
                            message = (e6 + '( '+str(l[3]) + ' <= '
                                       + left_string + ' <= ' + str(l[4]) + ' )')
                            break                                   
                    elif (l[2]=='checkrangestrict'):
                        if (len(l) < 5):
                            status = 5
                            message = e5
                            break                                                   
                        elif ((l[0] <= l[3]) | (l[0] >= l[4])):   #value<=min or value>=max
                            status = 6
                            message = (e6 + '( ' + str(l[3]) + ' < '
                                       + left_string + ' < ' + str(l[4])+' )')
                            break
                    elif (l[2]=='checkrangestrictmin'):
                        if (len(l)<5):
                            status = 5
                            message = e5
                            break                                                   
                        elif ( (l[0]<=l[3]) | (l[0]>l[4]) ):    #value<=min or value>max
                            status = 6
                            message = (e6 + '( ' + str(l[3]) + ' < '
                                        + left_string + ' <= ' + str(l[4]) + ' )')
                            break
                    elif (l[2] == 'checkrangestrictmax'):             
                        if (len(l) < 5):
                            status = 5
                            message = e5
                            break                                                   
                        elif ((l[0] < l[3]) | (l[0] >= l[4])):    #value<min or value>=max
                            status = 6
                            message = (e6 + '( ' + str(l[3]) + ' <= '
                                       + left_string + ' < ' + str(l[4]) + ' )')
                            break
                    elif (l[2] == 'boolean'):
                        if (l[0] not in range(2)):
                            status = 6
                            message = e6 + '( '+left_string+' must be 0 or 1 )'
                            break   
                    else:
                        status = 7
                        message = e7 + ' (\''+l[2]+'\')'
                        break
        f_input.close()

        return (status, message)


    def write_file(self, filename):
        """ Write configuration data to a file.

            This method creates an ASCII file with the name given and writes 
            to it variable names and values from the dictionary self.d
            File format is like the following
            variable1 = 123.45E-6

            variable2 = 7654.3E-2

            Parameters
            ----------

            filename:       a string containing the name of input data file

            Used data attributes
            --------------------

            self.d:         dictionary containing variables and associated values

            Returns
            -------

            (status, message)
            status:  0 = no error
                     1 = could not open file
                     2 = error during file writing (this *should* never happen...)
            message: a string containing an error message or 'Ok'
        """

        status  = 0
        message = OK    
        try:
            f_output = open(filename,'w')
        except IOError:
            status  = 1
            message = 'could not open file'
            return (status, message)

        # Write names and values to file
        for l in range(len(self.d)):
            item_list = list(self.d.items())
            variable = item_list[l][0]
            value    = item_list[l][1][0]
            comment  = '#' + item_list[l][1][1] + EOL
            if   (value == True):  value = 1 
            elif (value == False): value = 0
            string = variable + " = " + str(value) + EOL + EOL
            try:
                f_output.write(comment)
                f_output.write(string)
            except IOError:
                status = 2
                message = "line "+str(line)+": error while writing line"
                break

        f_output.close()

        return (status, message)


#+---------------+
#| Data sequence |
#+---------------+

class DataSequence:
    """ A monodimensonal sequence of values, to be read from a ASCII file """

    def read_file(self, filename, sep=EOL):
        """ Read a monodimensonal sequence of values from a ASCII file.

            filename:               a string containing the name of the file
            sep:                    character used as separator between values (usually the end-of-line character)

            Initialized data attributes
            ---------------------------

            self.data_array:        array of values
            self.n_data:            number of data in the array

            Returns
            -------

            (status, message)
            status:  0 = no error
                     1 = could not open file
                     2 = error reading file
                     3 = no data
            message: a string containing an error message or 'Ok'
        """

        status  = 0                             # This variable is used for error checking
        message = OK

        try:
            f_input = open(filename,'r')
        except IOError:
            status  = 1
            message = 'could not open file: "' + filename + '"'
            return (status, message)

        # Read data from input file
        try:
            self.data_array = numpy.fromfile(file=f_input, dtype=float, count=-1, sep=sep)
        except (IOError or ValueError):
            status  = 2
            message = 'error reading file: "' + filename + '"'
            return (status, message)                        
        self.n_data = len(self.data_array)
        if (self.n_data==0):
            status  = 3
            message = 'no valid data retrieved'
            return (status, message)

        return (status, message)


    def write_file(self, filename, sep=EOL):
        """ Write a monodimensional numpy array to a ASCII file.

                Parameters
                ----------

                filename:               a string containing the name of the file

                sep:                    character used as separator between values (usually the end-of-line character)

                Used data attributes
                --------------------

                self.data_array:        array of values

                Returns
                -------

                (status, message)
                status:  0 = no error
                         1 = could not open file
                         2 = error during file writing (this *should* never happen...)
                message: a string containing an error message or 'Ok'
        """

        status, message = 0, OK 

        # Open file for writing 
        try:
            f_output = open(filename,'w')
        except IOError:
            status  = 1
            message = 'could not open file: "' + filename + '"'
            return (status, message)

        self.data_array.tofile(file=f_output, sep=sep, format="%s")

        f_output.close()

        return (status, message)

       

#+-----------+
#| Data grid |
#+-----------+

class DataGrid:
    """ A 2-dimensional array of data.

        This class defines a 2-dimensional numpy into which a numerical grid (two dimensional matrix) 
        can be read from a ascii file.
    """


    def read_file(self, filename, n_rows=0, n_columns=0,
                  sep=SEP, pad_value=0, transpose=False, skip=0, debug=False):
        """ Read a 2-d array of values from file.

            Read a table of values from a ASCII file into a numpy bidimensional array
            Empty lines, as well as lines having '#' as first non-blank character are ignored

            filename:   a string containing the name of the ASCII file
            n_rows:     number of rows (0 = autodetect)
                        if it exceeds the number of rows in the data, 
                        the array is padded with the value defined by the pad_value parameter
            n_columns:  number of columns (0 = autodetect)
                        if it exceeds the number of columns in the data, 
                        the array is padded with with the value defined by the pad_value parameter
            sep:        character used as separator between columns of the table 
                        (usually '\t' or ',')
            pad_value:  value used to fill the empty cells of the grid
            transpose:  if set to 
                            False: the array will be of the form [row, column] 
                            True:  the array will be of the form [column, row]
            skip:       number of lines to skip at file beginning
                        useful to skip the header of some kind of data file formats
            debug:      if set to True, debug information is printed

            Initialized data attributes
            ---------------------------

            self.data_array:        2-d numpy array

            Returns
            -------

            (status, message)
            status:  0 = no error
                     1 = could not open file
                     2 = syntax error
                     3 = non-numerical value
                     4 = invalid number of rows/columns
                     5 = invalid number of rows to skip
            message: a string containing an error message or 'Ok'
        """

        status, message  = 0, OK                # This variable is used for error checking      

        # Check that the given numbers of rows ad columns are valid
        if (n_rows < 0) | (n_columns < 0):
            status  = 4
            message = 'invalid number of rows/columns'
            return (status, message)

        # Check that the number of rows to skip (if given) is positive
        if (skip < 0):
            status  = 5
            message = 'invalid number of rows to skip'
            return (status, message)

        # Open ASCII file
        try:
            f_input = open(filename,'r')
        except IOError:
            status  = 1
            message = 'could not open file'
            return (status, message)

        # If numbers of rows and/or columns are unknown, get them
        if (n_rows == 0) | (n_columns == 0):    # Get the number of rows and/or columns
            (rows, cols) = _get_rows_columns(f_input, sep, skip, debug)
            if debug: print("rows, cols = ", rows, cols)
            if (n_rows    == 0):    n_rows    = rows 
            if (n_columns == 0):    n_columns = cols                        

        # Read the data
        (status, message) = self._read_file(f_input, n_rows, n_columns,
                                            sep, pad_value, transpose, skip, debug)
        f_input.close()

        return (status, message)


    def _read_file(self, f_input, n_rows, n_columns, sep, pad_value, transpose,
                   skip, debug=False):
        """ Read a 2-d array of data from file (no check).

            Read a table of values from a ASCII file into a numpy bidimensional array
            WARNING: this method is not intended to be called directly: use read_file instead

            Parameters
            ----------

            filename:               a string containing the name of the file
            sep:                    character used as separator between columns of the table 
                                            (usually '\t' or ',')
            n_rows:                 number of rows
            n_columns:              number of columns
            transpose:              is set to 
                                            False: the array will be of the form [row, column] 
                                            True:  the array will be of the form [column, row]

            skip:                   number of lines to skip at file beginning
            debug:                  if True, debug information is printed

            Initialized data attributes
            ---------------------------

            self.data_array:        2-d numpy array

            Returns
            -------

            (status, message)
            status:  0 = no error
                     2 = syntax error
                     3 = non-numerical value
            message: a string containing an error message or 'Ok'

        """

        status, message  = 0, OK

        # Initialize numpy array
        if transpose:
            self.data_array = pad_value * numpy.ones((n_columns, n_rows))
        else:           
            self.data_array = pad_value * numpy.ones((n_rows, n_columns))

        line = 0        # Line of the ASCII file
        row  = 0        # Row of the table
        for string in f_input:
            line = line + 1
            if debug: print("Line: " +str(line)+ " = \'" +string+ "\'")
            if (line >= skip+1):
                # Cut newline charater from the end of line
                string = string.strip(EOL)
                # Cut line portion at right of '#' char
                string = string.split('#')[0]
                # Cut spaces from beginning and end of line
                string = string.strip()

                # Ignore line if it is empty
                if (len(string) != 0):
                    # Get a list of the strings in the line
                    l = string.split(sep)
                    for column in range(n_columns):
                        if (column <= len(l)-1):
                            # Try to transform string value to floating-point
                            try:
                                value = float(l[column].strip())
                            except ValueError:
                                status  = 3
                                message = "line "+str(line)+": non-numerical value"
                                break
                            if transpose:
                                self.data_array[column, row] = value
                            else:
                                self.data_array[row, column] = value
                    row = row + 1
                if (row > n_rows-1): break

        return (status, message)


    def write_file(self, filename, sep=SEP, transpose=False):
        """ Write a bidimensional numpy array to a ASCII file.

            Parameters
            ----------

            filename:               a string containing the name of the file
            sep:                    character used as separator between the two columns of the table 
                                    (usually <TAB> or ',')
            transpose:              is set to 
                                    * False: the array is saved in the form [row, column] 
                                    * True:  the array is saved in the form [column, row]
                                    where column is in range (0,1)

            Used data attributes
            --------------------

            self.data_array:        2-dimensional array of values

            Returns
            -------

            (status, message)
            status:  0 = no error
                     1 = could not open file
                     2 = table inconsistency
                     3 = error during file writing (this *should* never happen...)
            message: a string containing an error message or 'Ok'
        """

        status, message = 0, OK 

        # Open file for writing 
        try:
            f_output = open(filename,'w')
        except IOError:
            status  = 1
            message = 'could not open file'
            return (status, message)

        if transpose:   (n_columns, n_rows) = self.data_array.shape
        else:           (n_rows, n_columns) = self.data_array.shape

        # Write table to file
        for row in range(n_rows):
            string = ""
            for col in range(n_columns):
                if transpose: string = string + str(self.data_array[col, row])
                else:         string = string + str(self.data_array[row, col])
                if (col < n_columns-1): string = string + sep
            string = string + EOL
            try:
                f_output.write(string)
            except IOError:
                status = 2
                message = "line "+str(line)+": error while writing"
                break

        f_output.close()

        return (status, message)

#+-----------+
#| Utilities |
#+-----------+

def _get_rows_columns(f_input, sep=SEP, skip=0, debug=False):
    """ Gets number of rows and colums from a ascii data-file.

        Gets the number of rows and colums of matrix saved in a ASCII data file
        Empty lines, as well as lines starting with the `#' character are not considered

        Parameters
        ----------

        f_input:        data file identifier
        sep:            character used as separator between data inside rows
                        (usually '\t' or ',')
        skip:           number of ASCII lines to skip at the file beginning
                        this may be useful to skip the header of some kind of data file formats
        

        Returns
        -------

        (rows, columns)
    """

    line = 0
    rows = 0
    cols = 0
    for string in f_input:
        line = line + 1
        if (line >= skip+1):
            # Cut newline charater from the end of line
            string = string.strip(EOL)
            # Cut line portion at right of '#' char
            string = string.split('#')[0]
            # Cut spaces from beginning and end of line
            string = string.strip()
            if len(string)!=0:  
                rows = rows+1
                l = len(string.split(sep))
                if (l > cols): cols = l
    f_input.seek(0)
    
    return (rows, cols)





#!/usr/bin/python

# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This is a small demo script, to show the main features of this package.

    If the package has been correctly installed, you should be able to run 
    this script from the system terminal by typing::

    $ python -m gnuplot_manager.demo

    or inside the python console by typing::

    >>> from gnuplot_manager.demo import main
    >>> main()

    The demo will run without need of input and last about 35 seconds.    
"""



from time import sleep
import numpy
from .functions import *

def wait():
    sleep(3)
    
def main():

    x = numpy.linspace(0, 100, 51)

    print('\n***Demo of the gnuplot_manager package***\n')
    
    print('In this demo, gnuplot output is silenced\n')
    p2d = new_plot(xpos=600, ypos=100, width=600, height=450,
                   plot_type='2D', title='2D PLOT WINDOW DEMO', redirect_output=None)        
    p3d = new_plot(xpos=680, ypos=200, width=600, height=450,
                   plot_type='3D', title='3D PLOT WINDOW DEMO', redirect_output=None)    
    
    print(' 1) -> Plotting a parabola from data\n')
    plot2d(p2d, x, x*x, 'parabola')
    wait()

    print(' 2) -> Plotting a parabola as a function\n')
    plot_function(p2d, 'x**2', replot=True)
    wait()
    
    print(' 3) -> Changing the scale limits\n')
    plot_set(p2d, xmin=1, xmax=5, ymin=1, ymax=25, replot=True)
    wait()

    print(' 4) -> Setting logarithmic x and y axes\n')
    plot_set(p2d, logx=True, logy=True, replot=True)
    wait()

    print(' 5) -> Resetting ranges and scales\n')
    plot_set(p2d, resetx=True, resety=True, logx=False, logy=False, replot=True)
    wait()    

    print(' 6) -> Adding a label to the plot\n')
    plot_label(p2d, x=20, y=20, label='THIS IS A LABEL', replot=True)
    wait()

    print(' 7) -> Removing the label\n')
    plot_label(p2d, label=None, erase=True, replot=True)
    wait()

    print(' 8) -> Plotting another curve with volatile data\n')
    plot2d(p2d, x, x*x*x/100, 'x^3/100 (volatile)', volatile=True, replot=True)
    wait()            

    print(' 9) -> Plotting a 3D curve from data\n')
    plot3d(p3d, x, x, 2*x*x, '3D curve')
    wait()

    print('10) -> Plotting a 3D surface as a function\n')
    plot_function(p3d, 'x**2+y**2', replot=True)
    wait()

    print('11) -> Setting logarithmic z axis\n')
    plot_set(p3d, logz=True, replot=True)
    wait()

    print('12) -> Listing the plots\n')
    plot_list()
    wait()

    print('13) -> Closing the 2d plot\n')
    plot_close(p2d)
    wait()

    print('14) -> Listing the plots again')
    plot_list()
    wait()    

    print('15) -> Closing all the plots, removing files and directories\n')
    plot_close_all()

    print('***End of demo***')
    
if __name__ == '__main__':
    main()



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Classes for the simulation of neutral and charged particles in a plasma discharge.

    This module contains some classes defining the characteristics of reactors and the physical quantities involved
    in their usage, such as electric fields, electrodes dimensions and so on.

    Documentation is also available in the docstrings.
"""

# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import math
import numpy
from pysica.plasmapro.ccpla_defaults import *

# +-------------+
# | CCP Reactor |
# +-------------+

class CcpProperties:
        """"A class defining the physical properties of a capacitively coupled cold plasma reactor."""

        def __init__(self, d, l, V, f, phi, n_cells, lateral_loss):
                """Initialises the properties of a CCP plasma reactor.

                        Parameters
                        ----------

                        d:            distance between the electrodes
                        l:            lateral length the electrodes
                        V:            potential difference applied between the electrodes: peak value
                        f:            frequency of the bias applied between the electrodes
                        phi:          starting phase of the potential difference
                        n_cells:      number of cells used for PIC scheme
                        lateral_loss: if set to True, electrons and ions that reach the x or y borders will be lost,
                                      otherwise will be recovered


                        Initialized data attributes
                        ---------------------------

                        self.distance           distance between the electrodes
                        self.length             lateral length the electrodes
                        self.area               area of the electrodes
                        self.volume             plasma volume
                        self.lateral_loss:      if set to True, electrons and ions that reach the x or y borders will be lost,
                                                otherwise will be recovered
                        self.E_peak             intensity of the average electric between the electrodes: peak value
                        self.V_peak             intensity of the external electric bias applied between the electrodes: peak value
                        self.frequency          frequency of the external electric bias applied between the electrodes
                        self.pulsation          pulsation of the external electric bias applied applied between the electrodes
                        self.period             period of the external electric bias applied applied between the electrodes
                        self.phase              phase of the external electric field applied to the particles
                        self.average_current    average current measured at the electrodes 
                                                it is defined as a rank-zero numpy array in order to get working f2py 
                                                automatic conversion between fortran and python
                        self.potential          array used to store electric potential values at cells boundaries in PIC scheme
                        self.charge_density     array used to store electric charge density values at cells boundaries in PIC scheme
                        self.grid_points        array used to store the position along z-axis of the grid nodes in PIC scheme

                """
                
                self.distance           = d
                self.length             = l
                self.area               = l*l
                self.volume             = self.area * d
                self.lateral_loss       = lateral_loss
                self.E_peak             = V / d
                self.V_peak             = V 
                self.V                  = self.V_peak * math.sin(phi)
                self.frequency          = f
                self.pulsation          = 2 * math.pi * f
                if (f > 0):
                        self.period     = 1.0 / f
                else:
                        self.period     = numpy.inf
                self.phase              = phi
                self.average_current    = numpy.zeros(1, 'd')          # Needs to be an array since it will be passed to f2py
                self.potential          = numpy.zeros(n_cells+1, 'd')
                self.charge_density     = numpy.zeros(n_cells+1, 'd')
                self.grid_points        = numpy.zeros(n_cells+1, 'd')
                self.delta_grid         = d / n_cells
                for i in range(n_cells+1):
                        self.grid_points[i] = i * self.delta_grid


        def initialize_savefiles(self, filename_I, filename_V, append=False, sep=SEP, ext='.csv'):
                """ Initialize the file where the electric potential spatial distribution is saved

                    Parameters
                    ----------
                    filename_I:            name of the file to which the electric current at the electrodes is saved
                    filename_V:            name of the file to which the electric potential values are saved
                    sep:                   character or sequence used to separate data columns
                    ext:                   extension to give to the filename (default is '.csv')

                    Initialized data attributes
                    ---------------------------

                    append:                if True, open the file for appending new data to the end,
                                           the header will not be re-written
                    self.sep:              character or sequence to separate data columns
                    delf.f_I:              file to which values of electric current are saved
                    self.f_v:              file to which values of electric potential are saved
                """

                self.sep = sep
                self.filename_I = str(filename_I) + ext
                self.filename_V = str(filename_V) + ext                        

                # Open data file where to save electric current
                if append:
                        data_file_I = open(self.filename_I,'a')
                else:
                        data_file_I = open(self.filename_I,'w')
                        # Write header
                        data_file_I.write('\"t[ns]\"' + self.sep + 'I [A]')
                        data_file_I.write(EOL)
                data_file_I.close()

                # Open data file where to save electric potential spatial distribution
                if append:
                        data_file_V = open(self.filename_V,'a')
                else:
                        data_file_V = open(self.filename_V,'w')
                        # Write header
                        n = len(self.grid_points)
                        for i in range(n):
                                data_file_V.write(str(self.grid_points[i] * 1.0E-3))
                                if (i < n-1): data_file_V.write(self.sep)
                        data_file_V.write(EOL)
                data_file_V.close()


        def save_data_to_files(self, time):
                """Saves actual data values of electric current and potential spatial distribution to the savefiles """

                data_file_I = open(self.filename_I,'a')                        
                data_file_I.write( str(1E9*time) + self.sep + str(self.average_current[0]) )
                data_file_I.write(EOL)
                data_file_I.close()

                data_file_V = open(self.filename_V,'a')
                n = len(self.potential)
                for i in range(n):
                        data_file_V.write(str(self.potential[i]))
                        if (i < n-1): data_file_V.write(self.sep)
                data_file_V.write(EOL)
                data_file_V.close()



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Simulation of a capacitively coupled plasma discharge: functions used to plot data in real time """

# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import math
import numpy


# Import required constants and parameters
from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *
from pysica.managers.gnuplot_manager import *


#+---------------------------+
#| General purpose functions |
#+---------------------------+

def _decimate(a, k=MAX_DECIMATION_FACTOR):
    """ Keeps only an item every k ones from an array 

        Parameters
        ----------

        a:  original array
        k : decimation factor        

        Returns
        -------
        decimated array
    """
    
    if (k < 2): k=2
    d = numpy.zeros(len(a))
    n = int(len(a)/k)
    for i in range(n):
        d[i] = a[int(i*k)]
    return d,n


def decimate(a, n=MAX_DECIMATION_FACTOR):
    """ Removes an item every n ones from an array

        Parameters
        ----------

        a: original array
        n: decimation factor

        Returns
        -------
        decimated array
    """
    
    if (n < 2)       : n = 2
    if (n > len(a)-1): n = len(a)-1
    d = numpy.zeros(len(a))
    j = 0 # index on d
    k = 0 # decimation counter
    for i in range(1, len(a)):
        k += 1
        if (k < n):
            d[j] = a[i]
            j += 1                        
        else:
            k =  0
    return d, j



#+---------------+
#| Runtime plots |
#+---------------+
    
class CcplaPlots:
        
    def __init__(self, parameters, charges, ccp, screen_width=SCREEEN_WIDTH, screen_height=SCREEEN_HEIGHT,
                 vertical_shift=VERTICAL_SHIFT, hborder=HBORDER, vborder=VBORDER):
        """ Defines an istance of the class CcplaPlots, which contains all the runtime plots

            The ensamle of plots is initialized (dimensions, positions etc...)
        """
        
        # Define dimensions of plot windows
        self.window_width          = int( screen_width / 4 - hborder)
        self.window_height         = int( (screen_height - 2*vertical_shift) / 3 - vborder )
        self.vertical_shift        = vertical_shift
        self.hborder               = hborder
        self.vborder               = vborder
        #self.dot_points            = parameters.dot_points
        self.ccp                   = ccp
        if parameters.dot_points: self.plot_style = 'dots'
        else:                     self.plot_style = 'points'

        # Type of gnuplot terminal
        self.str_terminal          = DEFAULT_TERMINAL
        
        self.initialize_graphs()

        # Length of history arrays after which they are decimated
        self.history_limit           = parameters.n_max_points + 1
        # Decimation factor
        self.history_decimation      = parameters.decimation_factor

        # Define numpy arrays that will be used to plot average electron energies during simulation
        self.history_time            = numpy.zeros(self.history_limit,'d')      # Elapsed simulation time
        self.history_e_average       = numpy.zeros(self.history_limit,'d')      # Mean electrons kinetic energy
        self.history_n_electrons     = numpy.zeros(self.history_limit,'i')      # Number of electrons
        self.history_nw_electrons    = numpy.zeros(self.history_limit,'d')      # Number of electrons, weighted

        self.reset_history()

        
    def plot_time(self, graph, time):
        """ Prints simulation time on a graph 
             
            Parameters
            ----------
            graph:  graph on which the time must be printed
                    it must have been created using the new_plot function
            time:   value of the simulation time

            Returns
            -------
            graph_x: absolute position of the graph window in screen coordinates

        """        
        string_time = 't = ' + str(round(time*1E9,3)) + ' ns'
        plot_label(graph, x=LABEL_XPOS, y=LABEL_YPOS, label=string_time, erase=True)    

        
    def graph_x(self, x):
        """ Trasforms logical x coordinate in the absolute coordinate to use for graph positioning 
             
            Parameters
            ----------
            x:  x logical coordinate: 
                x=0 means first graph from the left, 
                x=1 means second graph from the left ....

            Returns
            -------
            graph_x: absolute position of the graph window in screen coordinates

        """
        return x * (self.window_width + self.hborder)

    
    def graph_y(self, y):
        """ Trasforms logical y coordinate in the absolute coordinate to use for graph positioning 
             
            Parameters
            ----------
            y:  y logical coordinate: 
                y=0 means first graph from the top, 
                y=1 means second graph from the top ....

            Returns
            -------
            graph_y: absolute position of the graph window in screen coordinates

        """        
        return self.vertical_shift + y *(self.window_height + self.vborder)
                
               
    def initialize_graphs(self, reset=False):
        """ Creates and initializes all the graphs 
              
            Parameters
            ----------

            reset: if set to True, 
                   - all open graphs are closed before opening the new ones
                   - the history is reset
        """

        # If requested, clears all open graphs and reset history
        if reset:
            # Close all open graphs
            plot_close_all()
            # Reset numpy arrays that will be used to plot average electron energies during simulation          
            self.reset_history()
           
        # Number of electrons vs time
        self.n_electrons_graph              = new_plot(xpos=self.graph_x(0), ypos=self.graph_y(0),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       title='Number of electrons',
                                                       xlabel='Time / ns', ylabel='Electrons',
                                                       xmin=0,
                                                       logy=True,
                                                       redirect_output=True,
                                                       purge=False)
        # Mean energy of electrons vs time
        self.electrons_average_energy_graph = new_plot(xpos=self.graph_x(1), ypos=self.graph_y(0),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       title='Electron mean energy',
                                                       xlabel='Time / ns', ylabel='Energy / eV',
                                                       redirect_output=True,
                                                       purge=False)
        # Charge density vs z-position
        self.charge_density_graph           = new_plot(xpos=self.graph_x(2), ypos=self.graph_y(0),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=0,
                                                       xmax=self.ccp.distance*1E3,
                                                       format_y='%.2g',
                                                       title='Charge density',
                                                       xlabel='z / mm', ylabel='Rho / C  m**-3',
                                                       redirect_output=True,
                                                       purge=False)
        # Electric potential vs z-position
        self.potential_graph                = new_plot(xpos=self.graph_x(3), ypos=self.graph_y(0),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=0,
                                                       xmax=self.ccp.distance*1E3,
                                                       title='Electric potential',
                                                       xlabel='z / mm', ylabel='Potential / V',
                                                       redirect_output=True,
                                                       purge=False)
        # Electron energy vs angle
        self.electrons_phase_space_graph    = new_plot(xpos=self.graph_x(0), ypos=self.graph_y(1),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=-5,
                                                       xmax=185,
                                                       format_y='%.2g',
                                                       style=self.plot_style,
                                                       title='Electron energy vs angle',
                                                       xlabel='Angle / deg', ylabel='Energy / eV',
                                                       redirect_output=True,
                                                       purge=False)
        # Ion energy vs angle
        self.ions_phase_space_graph         = new_plot(xpos=self.graph_x(1), ypos=self.graph_y(1),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=-5,
                                                       xmax=185,
                                                       format_y='%.2g',
                                                       style=self.plot_style,
                                                       title='Ion energy vs angle',
                                                       xlabel='Angle / deg', ylabel='Energy / eV',
                                                       redirect_output=True,
                                                       purge=False)
        # Electron energy vs z-position
        self.electrons_phase_space_graph2   = new_plot(xpos=self.graph_x(2), ypos=self.graph_y(1),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=0,
                                                       xmax=self.ccp.distance*1E3,
                                                       format_y='%#.2g',
                                                       style=self.plot_style,
                                                       title='Electron energy vs z-position',
                                                       xlabel='z / mm', ylabel='Energy / eV',
                                                       redirect_output=True,
                                                       purge=False)
        # Ion energy vs z-position
        self.ions_phase_space_graph2        = new_plot(xpos=self.graph_x(3), ypos=self.graph_y(1),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=0,
                                                       xmax=self.ccp.distance*1E3,
                                                       format_y='%#.2g',
                                                       style=self.plot_style,
                                                       title='Ion energy vs z-position',
                                                       xlabel='z / mm', ylabel='Energy / eV',
                                                       redirect_output=True,
                                                       purge=False)
        # Electron energy distribution
        self.eedf_graph                     = new_plot(xpos=self.graph_x(0), ypos=self.graph_y(2),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       style='histeps',
                                                       title='Electron energy distribution',
                                                       xlabel='Energy / eV', ylabel='Number of electrons / %',
                                                       redirect_output=True,
                                                       purge=False)
        # Ion energy distribution
        self.iedf_graph                     = new_plot(xpos=self.graph_x(1), ypos=self.graph_y(2),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       style='histeps',
                                                       title='Ion energy distribution',
                                                       xlabel='Energy / eV', ylabel='Number of ions / %',
                                                       redirect_output=True,
                                                       purge=False)
        # Position of electons in 3D
        self.electrons_position_graph       = new_plot(xpos=self.graph_x(2), ypos=self.graph_y(2),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=0, xmax=self.ccp.length*1E3,
                                                       ymin=0, ymax=self.ccp.length*1E3,
                                                       zmin=0, zmax=self.ccp.distance*1E3,
                                                       style=self.plot_style,
                                                       plot_type='3D',
                                                       title='Electron location',
                                                       xlabel='x / mm',
                                                       ylabel='y / mm',
                                                       zlabel='z / mm',
                                                       redirect_output=True,
                                                       purge=False)
        # Position of ions in 3D
        self.ions_position_graph            = new_plot(xpos=self.graph_x(3), ypos=self.graph_y(2),
                                                       terminal=self.str_terminal,
                                                       width=self.window_width,
                                                       height=self.window_height,
                                                       xmin=0, xmax=self.ccp.length*1E3,
                                                       ymin=0, ymax=self.ccp.length*1E3,
                                                       zmin=0, zmax=self.ccp.distance*1E3,
                                                       style=self.plot_style,
                                                       plot_type='3D',
                                                       title='Ion location',
                                                       xlabel='x / mm',
                                                       ylabel='y / mm',
                                                       zlabel='z / mm',
                                                       redirect_output=True,
                                                       purge=False)
        

    def graph_list(self, expanded=False):
        """ Return a list of all the plots """
        
        return plot_list(expanded=expanded, printout=False, getstring=True)

    def print_graphs(self, terminal, dirname=None, options=None):
        """ Export all the graphs to files """

        return plot_print_all(terminal, dirname, options)

    def reset_graphs(self, ccp, parameters):
        """ Reset graphs dimensions and other parameters """

        self.ccp = ccp
        if parameters.dot_points: self.plot_style = 'dots'
        else:                     self.plot_style = 'points'        
        
        # Charge density vs z-position
        plot_set(self.charge_density_graph,         xmin=0, xmax=self.ccp.distance*1E3)
                 
        # Electric potential vs z-position
        plot_set(self.potential_graph,              xmin=0, xmax=self.ccp.distance*1E3)

        # Electron energy vs angle
        plot_set(self.electrons_phase_space_graph,  style=self.plot_style)
        
        # Ion energy vs angle
        plot_set(self.ions_phase_space_graph,       style=self.plot_style)       

        # Electron energy vs z-position
        plot_set(self.electrons_phase_space_graph2, xmin=0, xmax=self.ccp.distance*1E3, style=self.plot_style)
        
        # Ion energy vs z-position
        plot_set(self.ions_phase_space_graph2,      xmin=0, xmax=self.ccp.distance*1E3, style=self.plot_style)

        # Position of electons in 3D
        plot_set(self.electrons_position_graph,     xmin=0, xmax=self.ccp.length*1E3,
                                                    ymin=0, ymax=self.ccp.length*1E3,
                                                    zmin=0, zmax=self.ccp.distance*1E3,
                                                    style=self.plot_style)
        # Position of ions in 3D
        plot_set(self.ions_position_graph,          xmin=0, xmax=self.ccp.length*1E3,
                                                    ymin=0, ymax=self.ccp.length*1E3,
                                                    zmin=0, zmax=self.ccp.distance*1E3,
                                                    style=self.plot_style)
        

    def clear_graphs(self):
        """ Clear all the graph windows """
        
        #plot_clear_all()
        plot_reset_all()        

    def raise_graphs(self):
        """ Raises all graph windows """

        plot_raise_all()


    def lower_graphs(self):
        """ Lowers all graph windows """

        plot_lower_all()
                                                  
    def reset_history(self):
        """ Reset to zero all values of history """
        
        self.i_history = 0
        self.history_time.fill(0)
        self.history_e_average.fill(0)
        self.history_n_electrons.fill(0)
        self.history_nw_electrons.fill(0)
                
    def update_history(self, charges, time):
        """ Updates the history array """
        
        if (self.i_history >= self.history_limit):
            (self.history_time,new_i) = decimate(self.history_time,         self.history_decimation)
            self.history_e_average    = decimate(self.history_e_average,    self.history_decimation)[0]
            self.history_n_electrons  = decimate(self.history_n_electrons,  self.history_decimation)[0]
            self.history_nw_electrons = decimate(self.history_nw_electrons, self.history_decimation)[0]
            self.i_history = new_i

        self.history_time[self.i_history]         = time
        self.history_e_average[self.i_history]    = charges.e_average(0)
        self.history_n_electrons[self.i_history]  = charges.n_active(0)
        self.history_nw_electrons[self.i_history] = charges.n_active(0) * charges.weight[0]
        self.i_history += 1

                                                       
    def plot_positions(self, time, charges):
        """ Plots positions of electrons and ions in 3D """

        # Update time on graphs
        self.plot_time(self.electrons_position_graph, time)                
        self.plot_time(self.ions_position_graph, time)
        
        # Plot electrons graphs
        plot3d(self.electrons_position_graph,
               charges.x[0][charges.active[0]]*1.0E3,
               charges.y[0][charges.active[0]]*1.0E3, 
               charges.z[0][charges.active[0]]*1.0E3,
               label=''+charges.names[0])

        # Plot ions graphs
        self.plot_list = []
        for i in range(1, charges.types):
            #print "Plotting ion type:"+str(i)+" (" + charges.names[i] + ") active n. "+str(charges.active[i].sum())
            # Plot only if there are active ions
            if (charges.n_active(i) > 0):
                self.plot_list.append( [ charges.x[i][charges.active[i]]*1.0E3, 
                                         charges.y[i][charges.active[i]]*1.0E3,
                                         charges.z[i][charges.active[i]]*1.0E3,
                                         charges.names[i], None ] )                
        plot_curves(self.ions_position_graph,self.plot_list)          
                

    def plot_phase_space_data(self, time, charges):
        """ Plots electron and ion energies vz angle and z-position """

        # Update time on graphs
        self.plot_time(self.electrons_phase_space_graph ,time)
        self.plot_time(self.electrons_phase_space_graph2 ,time)
        self.plot_time(self.ions_phase_space_graph, time)
        self.plot_time(self.ions_phase_space_graph2, time)
        
        # Plot electron energies vs angle and mean energy vs mean angle
        self.plot_list = []
        self.plot_list.append( [ numpy.degrees(charges.theta[0][charges.active[0]]),
                                 charges.energies(0),
                                 charges.names[0] + ' ', None ] )
        self.plot_list.append( [ numpy.array([numpy.degrees(charges.theta_average(0))]),
                                 numpy.array([charges.e_average(0)]),
                                 '<' + charges.names[0] + '>', None ] )
        plot_curves(self.electrons_phase_space_graph, self.plot_list)
        
        # Plot electron energies vs z-position
        plot2d(self.electrons_phase_space_graph2,
               charges.z[0][charges.active[0]]*1.0E3,
               charges.energies(0),
               label=charges.names[0])
        
        # Plot ion energy vs angle
        self.plot_list = []
        for i in range(1,charges.types):
            # Plot only if there are active ions
            if (charges.n_active(i) > 0):              
                self.plot_list.append( [ numpy.degrees(charges.theta[i][charges.active[i]]),
                                         charges.energies(i),
                                         charges.names[i], None ] )
        plot_curves(self.ions_phase_space_graph, self.plot_list)
            
        # Plot ion energy vs z-position
        self.plot_list = []
        for i in range(1,charges.types):
            # Plot only if there are active ions
            if (charges.n_active(i) > 0):              
                self.plot_list.append( [ charges.z[i][charges.active[i]]*1.0E3,
                                         charges.energies(i),
                                         charges.names[i], None ] )
        plot_curves(self.ions_phase_space_graph2, self.plot_list)

                                        
    def plot_distributions(self, time, charges):
        """ Plots electron and ion energy distributions """
        
        self.plot_time(self.eedf_graph, time)
        self.plot_time(self.iedf_graph, time)

        # Calculate the number of bins for electron distribution, as the square root of the number of electrons
        self.nbins_ele = int( math.sqrt( charges.n_active(0) ) )
        if (self.nbins_ele == 0): self.nbins_ele = 1
        if (self.nbins_ele > N_BINS_LOG): plot_set(self.eedf_graph, logx=True)
        else:                             plot_set(self.eedf_graph, logx=False) 

        # Plot EEDF
        if (charges.n_active(0) > 0):
            # Calculate bins and frequencies
            (h,b) = numpy.histogram( charges.energies(0), bins=self.nbins_ele,
                                    range=(0.0, charges.e_max(0)), density=False )
            # Remove the first item of the bins array (left value of the first bin)
            b = b[1:]
            # Calculate the bins width
            d = ( b.max() - b.min() ) / len(b)
            # Create an array with the center point of each bin
            b = b - d / 2.0
            # Transform from absolute frequencies (number of particles in each bin) to relative frequencies
            # (percentage of particles in each bin)
            h = h*100.0 / ( charges.n_active(0) * 1.0 )
            plot2d(self.eedf_graph, b, h, label=charges.names[0])
        
        # Calculate the number of bins for ion distribution, as the square root of the minimum number of ions
        # (all ion types)
        self.n = charges.n
        for i in range(1, charges.types):
            if (charges.n_active(i) < self.n): self.n = charges.n_active(i)
        self.nbins_ion = int(math.sqrt(self.n))
        if (self.nbins_ion == 0): self.nbins_ion = 1
        if (self.nbins_ion > N_BINS_LOG): plot_set(self.iedf_graph, logx=True)
        else:                             plot_set(self.iedf_graph, logx=False)
        
        # Set x-axis limit for ion graphs to max ion energy (all ion types)
        self.ion_emax = 0.0
        for i in range(1,charges.types):
            if (charges.n_active(i) > 0):
                if (charges.e_max(i) > self.ion_emax): self.ion_emax = charges.e_max(i)
        if (self.ion_emax > 0.0): plot_set(self.iedf_graph, xmax=self.ion_emax)

        # Plot IEDF for each ion type
        self.plot_list = []
        for i in range(1, charges.types):
            if (charges.n_active(i) > 0):
                # Calculate bins and frequencies
                (h,b) = numpy.histogram( charges.energies(i), bins=self.nbins_ion,
                                         range=(0.0, self.ion_emax), density=False )
                # Remove the first item of the bins array (left value of the first bin)
                b = b[1:]
                # Calculate the bins width
                d = ( b.max() - b.min() ) / len(b)
                # Create an array with the center point of each bin
                b = b - d / 2.0
                # Transform from absolute frequencies (number of particles in each bin) to relative frequencies
                # (percentage of particles in each bin)
                h = h*100.0 / ( charges.n_active(i) * 1.0 )
                self.plot_list.append( [b, h, charges.names[i], None] )
        plot_curves(self.iedf_graph, self.plot_list)
                    
                                        
    def plot_n_and_energy_data(self, time):
        """ Plot mean electron enery and number of electrons vs simulation time  """
        
        self.plot_time(self.electrons_average_energy_graph, time)
        self.plot_time(self.n_electrons_graph, time)

        # Rescale x-scale to start from min value left after decimation
        self.xmin = self.history_time[0]*1.0E9
        self.xmax = self.history_time[self.i_history-1]*1.0E9

        # Plot mean energy vs time
        plot_set(self.electrons_average_energy_graph, xmin=self.xmin, xmax=self.xmax)
        plot2d(self.electrons_average_energy_graph, self.history_time*1.0E9, self.history_e_average) 
                                                       
        #Plot number of computational and real (weighted) electrons                                                       
        plot_set(self.n_electrons_graph, xmin=self.xmin, xmax=self.xmax)
        self.plot_list = [ [self.history_time*1.0E9, self.history_n_electrons,  'comp e-', None],
                           [self.history_time*1.0E9, self.history_nw_electrons, 'real e-', None] ]
        plot_curves(self.n_electrons_graph,self.plot_list) 

        

    def plot_potential(self, time, ccp):
        """ Plots charge densiy and elecgric potential """

        self.plot_time(self.charge_density_graph, time)
        self.plot_time(self.potential_graph, time)
        plot2d(self.potential_graph,      ccp.grid_points*1.0E3, ccp.potential)
        plot2d(self.charge_density_graph, ccp.grid_points*1.0E3, ccp.charge_density)


#+----------------------+
#| Cross sections plots |
#+----------------------+

def plot_cross_sections(neutrals, electrons=True, ions=True, recombination=False, plot_all=False, dot_points=False):
    """ Plots cross sections """

    (status,message) = (0, OK)
    
    if electrons:
        (status, message) = neutrals.plot_xsec_electrons( plot_single      = not plot_all,
                                                          plot_total       = not plot_all, 
                                                          plot_frequencies = plot_all, 
                                                          plot_relative    = plot_all, 
                                                          plot_boundaries  = plot_all,
                                                          dot_points       = dot_points )
        if (status != 0): return (status, message)
                
    if ions:
        for ion_type in range(neutrals.types):
            (status, message) = neutrals.plot_xsec_ions( ion_type         = ion_type, 
                                                         plot_total       = not plot_all, 
                                                         plot_frequencies = plot_all, 
                                                         plot_relative    = plot_all, 
                                                         dot_points       = dot_points )
            if (status != 0): return (status, message)

    if recombination:                 
            (status, message) = neutrals.plot_xsec_ele_ion_recomb( plot_total = True, 
                                                                   plot_rates = plot_all,
                                                                   dot_points = dot_points )
   
    return (status, message)



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PYthon tools for SImulation and CAlculus: utility functions

    This subpackage contains some modules useful to manage the reding or writing of raw data or configuration variablees

    This sub-package contains some modules which can be used for managing data input/output from/to files, 
    print of physical quantities whith automatic managing of the units prefixes

    * subpackages
            - gnuplot manager        -> plot graphs by means of gnuplot

    * modules
            - data_manager           -> read/write data from/to ASCII files
            - unit_manager           -> prints physical quantities, managing the unit prefixes 
                                        (e.g. 1.23456E-9 A -> 1.23 nA)

    Documentation is also available in the docstrings.
"""



#!/usr/bin/python

# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module contains functions to plot data on 2D and 3D graphs by means of gnuplot.

This package allows to plot data or mathematical expressions inside python,
using the gnuplot program, in the form of 2D or 3D plots.

.. note:: If the package has been correctly installed, a small demo script can
   be run from the system terminal by typing::

   $ python -m gnuplot_manager.demo

   or inside the python console by typing::

   >>> from gnuplot_manager.demo import main
   >>> main()

   The demo will run without need of input and last about 35 seconds.

Full documentation can be found at:

https://github.com/pietromandracci/gnuplot_manager



List of available functions
===========================

Read the doctrings for a complete description of each function.

Create, modify, and close plot windows
--------------------------------------

*new_plot()*
    create a new plot window
*plot_set()*
    modify some properties of a previously created window
*plot_command()*
    send a command to the gnuplot process
*plot_close()*
    close the plot window and terminate the gnuplot process
*plot_close_all()*
    close all the plot windows and terminate all the gnuplot processes 


Plot data
---------

*plot1d()*
    plot a curve from 1d data
*plot2d()*
    plot a curve from 2d data
*plot3d()*
    plot a curve from 3d data
*plot_box()*
    plot a boxplot from 1d data
*plot_curves()*
    plot several curves at the same time

Plot mathematical functions
---------------------------

*plot_function()*
    plot a mathematical expression
*plot_functions()*
    plot several mathematical expression at once


Print plots to files
--------------------

*plot_print()*
    export a plot to a file

*plot_print_all()*
    export to files all the open plots


Reset, clear and refresh plots
------------------------------

*plot_reset()*
    reset a plot: remove all curves and functions
    and the clear the window 
*plot_reset_all()*
    reset all plot windows
*plot_clear()*
    clear the plot area
*plot_clear_all()*
    clear the plot area of all plots
*plot_replot()*
    refresh the plot window
*plot_replot_all()*
    refresh all the plot windows


Utility functions
-----------------

*plot_label()*
    print a string on the plot
*plot_raise()*
    rise the plot window over the other windows on the screen
*plot_lower()*
    lower the plot window under the other windows on the screen
*plot_raise_all()*
    rise all the plot windows    
*plot_lower_all()*
    lower all the plot windows
*plot_check()*
    print the plot properties
*plot_list()*
    print the properties of all plots


The *_PlotWindow* class
=======================

Each plot window is an instance of the *_PlotWindow* class, 
which has several attributes:

*self.window_number*:   
    an integer number that identifies the plot window, [#window_number]_                               
    mainly used to generate unique names for the data files
*self.gnuplot_process*: 
     gnuplot process (instance of *subprocess.Popen*)    
*self.term_type*:
    the type of gnuplot terminal    
*self.plot_type*:
    a string defining the type of plot : '2D', '3D',
    or *None* if the plot window has been closed
*self.n_axes:*
    number of plot axes (2 for 2D plots, 3 for 3D ones)
*self.xmin*:
    minimum of the x-axis (*None* if not set)
*self.xmax*:
    maximum of the x-axis (*None* if not set)
*self.ymin*:
    minimum of the y-axis (*None* if not set)
*self.ymax*:
    maximum of the y-axis (*None* if not set)
*self.zmin*:
    minimum of the z-axis (*None* if not set)
*self.zmax*:
    maximum of the z-axis (*None* if not set)
*self.persistence*:
    *True* if the plot was opened as persistent
*self.title*:
    the window title (*None* if not given)
*self.filename_out*: 
     name of the file to which gnuplot output is redirected
*self.filename_err*:
     name of the file to which gnuplot errors are redirected     
*self.data_filenames*:
     list containing the names of the datafiles related to the
     curves presently plotted on the window
*self.n_volatiles*:
     number of curves that have been plotted using the *volatile=True*
     argument: they are not listed in *self.data_filenames* since
     there are no associated data files
*self.functions*:
     list containing the function strings [#functions]_
*slef.purge*:
     if True, old data files are removed when new data is plotted
     without the *replot=True* option
*self.error*:
     if there was an error while opening the plot
     an error message is stored here

.. [#window_number] Note that this number is *not* the index that identifies the
   plot window inside the *window_list* variable: in fact the former is fixed,
   while the latter may change when other windows are removed from the list.

.. [#functions] Note that no check is made that function strings given to gnuplot 
   are correct. So even wrong ones (which therefore gnuplot has not plotted)
   are listed here.

.. note:: If you modify the plot by sending commands to gnuplot directly, using
   the *plot_command()* function, some of these attributes, such as the number of curves 
   and the list of data files, may not be updated properly.

The *_PlotWindow* class have some methods also, which are called by the functions
of the *functions.py* module to perform their tasks:

*self._command()*
    method used to send commands to gnuplot
*self._quit_gnuplot()*
    method used to terminate the gnuplot process and close the window
*self._add_functions()*
    method used to add one or more mathematical expression
*self._add_curves()*
    method used to add one or more curves from data

.. note:: Since the package is designed to use the functions in the
   *functions.py* module, these methods are not intended to be called directly.
"""

from .global_variables import *
from .functions import *




